'''
nicknames - determine if names are related, e.g., nicknames for each other

adapted from https://github.com/carltonnorthern/nickname-and-diminutive-names-lookup
'''
import collections
import csv
from os.path import join, dirname, abspath

class NameDenormalizer(object):
    '''
    denormalize any name

    usage:
        $ nn = NameDenormalizer()
        $ nn.get('jeff')
        {'geoff', 'jefferson', 'jeffrey', 'jefferey', 'geoffrey', 'sonny'}
    '''
    def __init__(self, filename=None):
        filename = filename or join(dirname(abspath(__file__)), 'nicknames.csv')
        lookup = collections.defaultdict(list)
        with open(filename) as f:
            reader = csv.reader(f)
            for line in reader:
                matches = set(line)
                for match in matches:
                    lookup[match].append(matches)
        self.lookup = lookup

    def __getitem__(self, name):
        name = name.lower()
        if name not in self.lookup:
            raise KeyError(name)
        names = set().union(*self.lookup[name])
        if name in names:
            names.remove(name)
        return names

    def get(self, name, default=None):
        '''
        get a set of names which work for the indicated name

        :param name: name to get set for
        :param default: return this if none found (default None)
        :return: set of names, or default
        '''
        try:
            return self[name]
        except KeyError:
            return default


###########################################################################################
#       Date            Author          Reason
#       ----            ------          ------
#       07/03/18        Lou King        Create
#
#   Copyright 2018 Lou King.  All rights reserved
###########################################################################################
'''
test_sqlalchemy_helpers  -- test sqlalchemy_helpers
=====================================================

'''
# standard
import unittest
from random import randint
from copy import deepcopy

# pypi
from faker import Faker
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

# home grown
from .models import Base, User, Address, NotUnique, SeveralAttrs
from loutilities.sqlalchemy_helpers import update, getunique, insert_or_update, dbConsistencyError, parameterError

TEST_DB = 'test.db'

class SqlalchemyHelpersTest(unittest.TestCase):

    # executed prior to each test
    def setUp(self):
        """Set up fake database session before all tests."""
        engine = create_engine('sqlite://', echo=False)  # echo=True for debug
        Base.metadata.create_all(engine)
        Session = sessionmaker(bind=engine)
        self.session = Session()

    # executed after each test
    def tearDown(self):
        pass

    ###############
    ### helpers ###
    ###############
    def populate_users(self, numUsers):
        """Create numUsers in fake database."""
        users = []
        addrs = []
        f = Faker()

        for i in range(numUsers):
            user, addr = self.create_user(f.name(), f.address())
            users.append(user)
            addrs.append(addr)

        self.session.add_all(users)
        self.session.commit()

        return users, addrs

    def create_user(self, name, address):
        """Create a fake user."""
        addr = Address(description=address)

        user = User(name=name, address=addr)

        return user, addr

    def populate_notunique(self, numValues):
        f = Faker()
        values = []
        for i in range(numValues):
            value = f.name()
            thisvalue = self.add_notunique(value)
            values.append(thisvalue)

        self.session.add_all(values)
        self.session.commit()

        return values

    def add_notunique(self, value):
        notunique = NotUnique(value=value)
        return notunique

    def populate_severalattrs(self, numRecs):
        f = Faker()
        records = []
        for i in range(numRecs):
            record = self.make_severalattrs()
            records.append(record)
        copyrecords = deepcopy(records)
        self.session.add_all(records)
        self.session.commit()

        return copyrecords

    def make_severalattrs(self):
        f = Faker()
        record = SeveralAttrs(intAttr1=randint(25, 42),
                              strAttr2=f.name(),
                              strAttr3=f.ssn(),
                              boolAttr4=bool(randint(0,1) == 1),
                              dateAttr5=f.date_time_this_decade()
            )
        return record

    ###############
    #### tests ####
    ###############

    def test_getunique_single(self):
        users, addrs = self.populate_users(3)

        thisuser = getunique(self.session, User, name=users[1].name)

        self.assertEqual(thisuser.name, users[1].name)

    def test_getunique_multiple(self):

        values = self.populate_notunique(3)
        notunique = self.add_notunique(values[1].value)
        self.session.add(notunique)
        self.session.commit()

        errormsg = ''
        try:
            thisuser = getunique(self.session, NotUnique, value=values[1].value)
        except dbConsistencyError as e:
            errormsg = str(e)

        self.assertIn('found multiple rows', errormsg)

    def test_getunique_none(self):

        values = self.populate_notunique(3)

        thisuser = getunique(self.session, NotUnique, value='nonesense')

        self.assertEqual(thisuser, None)

    def test_update(self):
        records = self.populate_severalattrs(5)

        oldinst = self.session.query(SeveralAttrs).filter_by(strAttr2=records[3].strAttr2).first()

        f = Faker()
        newattr3 = f.name()
        newattr5 = f.date_time_this_decade()
        newinst = SeveralAttrs(intAttr1=oldinst.intAttr1,
                               strAttr2=oldinst.strAttr2,
                               strAttr3=newattr3,
                               boolAttr4=oldinst.boolAttr4,
                               dateAttr5=newattr5
                )
        updated = update(self.session, SeveralAttrs, oldinst, newinst, skipcolumns=['id'])
        self.session.commit()

        self.assertEqual(updated, True)

        # check all but id field and sqlalchemy instance state field
        currinst = self.session.query(SeveralAttrs).filter_by(strAttr2=records[3].strAttr2).first()
        del currinst.id
        del currinst._sa_instance_state
        del newinst._sa_instance_state
        self.assertEquals(newinst.__dict__, currinst.__dict__)

    def test_insert_or_update_insert(self):
        oldrecords = self.populate_severalattrs(5)

        record = self.make_severalattrs()
        updated = insert_or_update(self.session, SeveralAttrs, record, skipcolumns=['id'], strAttr2=record.strAttr2)

        self.assertEqual(updated, True)

        newrecords = self.session.query(SeveralAttrs).all()
        self.assertEqual(len(newrecords), 6)

    def test_insert_or_update_update(self):
        oldrecords = self.populate_severalattrs(5)
        oldrecord = oldrecords[3]

        record = self.make_severalattrs()
        newinst = deepcopy(oldrecord)
        newinst.intAttr1 = record.intAttr1
        newinst.strAttr3 = record.strAttr3

        updated = insert_or_update(self.session, SeveralAttrs, newinst, skipcolumns=['id'], strAttr2=oldrecord.strAttr2)

        self.assertEqual(updated, True)

        newrecords = self.session.query(SeveralAttrs).all()
        self.assertEqual(len(newrecords), 5)

        currinst = self.session.query(SeveralAttrs).filter_by(strAttr2=oldrecord.strAttr2).first()
        del currinst.id
        del currinst._sa_instance_state
        del newinst._sa_instance_state
        self.assertEquals(newinst.__dict__, currinst.__dict__)



# ###############################################################################
# kmlutils - utilities for access to kml file
#
# REVISION HISTORY:
#	02/10/11    L King      Create
#
#   Copyright 2012 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#  
#
# ###############################################################################
"""
kmlutils - utilities for access to kml file
=============================================================

Provides methods to write KML files
"""

# standard
from xml.dom.minidom import getDOMImplementation as domimpl

# home grown
# NONE

class invalidChildattrs(Exception): pass

# ###############################################################################
class kmldoc():
# ###############################################################################
    """
    Represents kml document
    
    :param name: name of high level document
    :rtype: kmldoc instance
    """

    # ###############################################################################
    def __init__(self, name):
    # ###############################################################################
    
        self.doc = domimpl().createDocument ("http://www.opengis.net/kml/2.2","kml",None)
        self.kml = self.doc.documentElement
        self.kml.setAttribute ("xmlns","http://www.opengis.net/kml/2.2")
        self.name = name
        ne = self.doc.createElement("name")
        ne.appendChild (self.doc.createTextNode(self.name))
        self.kml.appendChild(ne)
        self.dd = self.doc.createElement('Document')
        self.kml.appendChild(self.dd)
    
    # ###############################################################################
    def save(self,filename):
    # ###############################################################################
        """
        save document in the indicated filename
        
        :param filename: name of file to save kml output to
        """
        
        OUT = open(filename,'w')
        self.doc.writexml(OUT,addindent='  ',newl='\n')

    # ###############################################################################
    def appendChildAttrs(self, element, attrs):
    # ###############################################################################
        """
        recursively append the attributes indicated by attrs to the element
        
        :param element: element to append children to
        :param attrs: ellist | attrtree
        
            * attrtree - dict with keywords for element's child attributes (may be nested)
            * ellist - list of elements to be appended, can have embedded attrtrees
            * Note: attrtree can contain embedded ellists or attrtrees
        """
        
        if isinstance(attrs, dict):
            for attr in attrs:
                atel = self.doc.createElement(attr)
                
                # dict means that this has attr name/value pairs which needs to be appended
                if isinstance(attrs[attr], dict):
                    self.appendChildAttrs(atel, attrs[attr])
                
                # list might have dict as above, or elements, which need to be appended
                elif isinstance(attrs[attr], list):
                    for elattr in attrs[attr]:
                        if isinstance(elattr, dict):
                            self.appendChildAttrs(atel, elattr)
                        else:
                            atel.appendChild(elattr)
                
                # otherwise we're assuming this is just a text node
                else:
                    atel.appendChild(self.doc.createTextNode(attrs[attr]))
                element.appendChild(atel)
        
        elif isinstance(attrs, list):
            for elattr in attrs:
                if isinstance(elattr, dict):
                    self.appendChildAttrs(element, elattr)
                else:
                    element.appendChild(elattr)
                    
        else:
            raise invalidChildattrs
        
    # ###############################################################################
    def namedel(self, name, attrs, childattrs):
    # ###############################################################################
        """
        return a named element with initial attributes
        
        :param name: tagname for the element
        :param attrs: dictionary with keywords for element's attributes
        :param childattrs: ellist | attrtree
        
            * attrtree - dict with keywords for element's child attributes (may be nested)
            * ellist - list of elements to be appended, can have embedded attrtrees
            * Note: attrtree can contain embedded ellists or attrtrees
        """

        element = self.doc.createElement(name)
        for attr in attrs:
            element.setAttribute(attr, attrs[attr])
        self.appendChildAttrs(element, childattrs)
        return element

# ###############################################################################
class style():
# ###############################################################################
    """
    represents a style
    
    :param kml: kml instance
    :param name: name of style
    :param childattrs: dictionary with keywords for style attributes (may be nested)
    """
    
    # ###############################################################################
    def __init__(self, kml, name, childattrs):
    # ###############################################################################
    
        self.kml = kml
        self.name = name
        self.childattrs = childattrs

    # ###############################################################################
    def el(self):
    # ###############################################################################
        """
        return kml element for this object
        """
    
        element = self.kml.doc.createElement('Style')
        element.setAttribute('id',self.name)
        self.kml.appendChildAttrs(element, self.childattrs)
        return element
    
    # ###############################################################################
    def elurl(self):
    # ###############################################################################

        element = self.kml.doc.createElement('styleUrl')
        element.appendChild(self.kml.doc.createTextNode ('#{0}'.format(self.name)))
        return element
        
# ###############################################################################
class coordinates():
# ###############################################################################
    """
    represents a list of coordinates
    
    :param kml: kml instance
    :param clist: list of coordinates
    """
    
    # ###############################################################################
    def __init__(self, kml, clist):
    # ###############################################################################
    
        self.kml = kml
        self.clist = clist

    # ###############################################################################
    def el(self):
    # ###############################################################################
        """
        return kml element for this object
        """
    
        element = self.kml.doc.createElement('coordinates')
        element.appendChild (self.kml.doc.createTextNode (" ".join([c.str() for c in self.clist])))
        return element
        
# ###############################################################################
class coordinate():
# ###############################################################################
    """
    represents a coordinate
    
    :param lat: latitude (decimal degrees)
    :param long: longitude (decimal degrees)
    :param alt: altitude (meters) 
    """

    # ###############################################################################
    def __init__(self, lat, long, alt=None):
    # ###############################################################################
    
        self.lat = lat
        self.long = long
        self.alt = alt

    # ###############################################################################
    def str(self):
    # ###############################################################################
        """
        return kml string for this coordinate
        """
    
        if self.alt != None:
            return '{0},{1},{2}'.format(self.long, self.lat, self.alt)
        else:
            return '{0},{1}'.format(self.long, self.lat)
            

# ###############################################################################
def main():
# ###############################################################################

    kml = kmldoc('test doc')
    
    styred = style(kml, 'red-outline', 
        {'LineStyle':{'color':'ff0000ff'}, 
        'PolyStyle':{'fill':'0'},
        })
    
    styblue = style(kml, 'blue-outline', 
        {'LineStyle':{'color':'ffff0000'}, 
        'PolyStyle':{'fill':'0'},
        })
    
    stygreen = style(kml, 'green-outline', 
        {'LineStyle':{'color':'ff00ff00'}, 
        'PolyStyle':{'fill':'0'},
        })
    
    kml.dd.appendChild(styred.el())
    kml.dd.appendChild(styblue.el())
    kml.dd.appendChild(stygreen.el())
    
    cl = []
    for i in range(30):
        cl.append(coordinate(i,i+1,i+2))
    cs = coordinates(kml,cl)
    kml.dd.appendChild(cs.el())
    
    place = kml.namedel('Placemark',attrs={'id':'poly'},childattrs=[{'name':'poly'},stygreen.elurl(),{'Polygon':{'outerBoundaryIs':{'LinearRing':[cs.el()]}}}] )
    kml.dd.appendChild(place)
    
    kml.save('kmlutilstest.kml')

# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    main()



# from https://gist.github.com/techniq/5174410
from datetime import datetime
from sqlalchemy import Column, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from flask_security import current_user


class AuditMixin(object):
    created_at = Column(DateTime, default=datetime.now)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)

    # @declared_attr
    # def created_by_id(cls):
    #     return Column(Integer,
    #         ForeignKey('user.id', name='fk_%s_created_by_id' % cls.__name__, use_alter=True),
    #         # nullable=False,
    #         default=_current_user_id_or_none
    #     )
    #
    # @declared_attr
    # def created_by(cls):
    #     return relationship(
    #         'User',
    #         primaryjoin='User.id == %s.created_by_id' % cls.__name__,
    #         remote_side='User.id'
    #     )
    #
    # @declared_attr
    # def updated_by_id(cls):
    #     return Column(Integer,
    #         ForeignKey('user.id', name='fk_%s_updated_by_id' % cls.__name__, use_alter=True),
    #         # nullable=False,
    #         default=_current_user_id_or_none,
    #         onupdate=_current_user_id_or_none
    #     )
    #
    # @declared_attr
    # def updated_by(cls):
    #     return relationship(
    #         'User',
    #         primaryjoin='User.id == %s.updated_by_id' % cls.__name__,
    #         remote_side='User.id'
    #     )


def _current_user_id_or_none():
    try:
        return current_user.id
    except:
        return None


################################################################################
# bfile - binary file handler
#   supports one file, can be opened for input or output
#
#   Date        Author      Reason
#   ----        ------      ------
#   09/24/09    L King      Create
#   10/07/09    L King      Add seek method
#   11/29/10    L King      Add brecord class
#   12/16/10    L King      Allow optional rangecheck within brecord class
#   01/05/11    L King      - In brecord class, add ability to specify expressions 
#                             in the "countfield" for repeated records
#                           - Handle unexpected end of file more gracefully
#   01/27/11    L King      instantiate recursive handler correctly based in current 
#                           handler's input parameters
#   01/28/11    L King      add getcountfields, getfieldorder methods to brecord class
#   01/04/12    L King      add byteorder parameter to bfile class
#
#   Copyright 2013 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
################################################################################
"""
bfile -- binary file handling
====================================

An object of this class supports one file, which can be opened for input or output.  Current implementation assumes file contains big-endian data.

Example use of :class:`bfile.bfile` opens a file, then reads a 2 byte unsigned field, then a 4 byte unsigned field::

    foo = bfile.bfile()
    foo.open(foofile, "rb")
    
    barnone     = foo.uget(2)
    barsome	= foo.uget(4)

    foo.close()
    
Example use of :class:`bfile.brecord` (note highest level is :attr:`foostruct` dictionary)::
    
    import bfile
    
    # need to create the lower level dictionaries first, else python will complain

    # make a convenience copy of the bfile.brecord sign indicators
    (U, TS, S, F) = (bfile.U, bfile.TS, bfile.S, bfile.F)

    # dictionary which describes file structure has scalars such as 'BARNONE', and subrecords such as 
    # 'satcontactrecs'.  Note use of tuples (e.g., "(0, 2, U)", with parens) for scalars, and lists 
    # (e.g., "[9,'NUM_SAT_CONTACTS', satcontactcomm, None]" with square brackets) for subrecords.  Use of list is 
    # required for subrecords because the data within is changed by bfile.brecord itself, and tuples are immutable
    
    # note range field is optional, and it must be a single entry 
    # to verify a min,max range, use the tuple value (min,max) -- **tuples (in parens) are assumed to be min,max!!**
    # to verify within a discrete set of values, use a list, e.g., [1,2,5] or range(6)
    
    foo                     = dict(#   ndx   len sign    range
        BARNONE             = (0,   1,  U,      (0,15)),
        SAT_PN              = (1,   1,  U,      (0,15)),
        ANT_REF_NUM         = (2,   1,  U,      (0,3)),
        T_START_TIME        = (3,   4,  TS),    # note no range specified here
        T_STOP_TIME         = (4,   4,  TS),
        CONTACT_ENRGY       = (5,   4,  U,      (0,480000)),
        NUM_SAT_PWR         = (6,   1,  U,      (1,20)),
                         #    ndx   count           structure       handler(added later)
        satpwrsegs          = [7,   'NUM_SAT_PWR',  satpwrsegcomm,  None],
        NUM_RHPOL_PWR       = (8,   1,  U),
        rhpolpwrsegrecs     = [9,   'NUM_RHPOL_PWR', polpwrsegcomm, None],
        NUM_LHPOL_PWR       = (10,  1,  U),
        lhpolpwrsegrecs     = [11,  'NUM_LHPOL_PWR', polpwrsegcomm, None],
        NUM_BEAMS_F         = (12,  1,  U),
        fwdbeamrecs         = [13,  'NUM_BEAMS_F',  fwdbeamreccomm, None],
        NUM_BEAMS_R         = (14,  1,  U),
        rvsbeamrecs         = [15,  'NUM_BEAMS_R',  rvsbeamreccomm, None],
        )

    raistruct = dict(      #  ndx  len sign     range
        GATEWAY_ID          = (0,   2,  U,      (1,100)),
        START_TIME          = (1,   4,  TS),
        STOP_TIME           = (2,   4,  TS),
        NUM_ALERTING        = (3,   1,  U,      [0,1,2,4]),
        ACCESS_CHAN         = (4,   2,  U),
        REV_ANCHOR_FREQ     = (5,   2,  U,      (4,496)),
        BACH_PWR_OFFSET     = (6,   2,  U,      range(100,900,25)),
        NUM_PAGING          = (7,   1,  U,      [2,4,8]),
        NUM_SAT_CONTACTS    = (8,   2,  U,      (1,512)),
                         #    ndx   count           structure       handler(added later)
        satcontactrecs      = [9,   'NUM_SAT_CONTACTS', foo, None],
        )
    
    # create the handler object
    raih = bfile.brecord(raistruct)


"""

# standard


import re
import struct
import sys


class invalidRangeCheck(Exception): pass

# ###############################################################################
def rangecheck (start,  fieldname, value, *range, **kwargs):
# ###############################################################################
    """
    range check data
    
    :param start: number of bytes into the file at the start of the field
    :param fieldname: name of the field
    :param value: value of the field
    :param range: verify value is within this range
    
        * If two values supplied, range[0] is min, range[1] is max
        * If tuple supplied, (min,max) is assumed
        * If list supplied, valid values are in the list
        
    :param rangeerr=sys.stdout: file to send output to
    """
    
    maxwidth = 30
    
    # default for rangeerr keyword arg is sys.stdout
    rangeerr = kwargs.get('rangeerr', sys.stdout)
    
    # save original parameters for possible exception
    origrange = range
    
    # if one value supplied, might be a tuple.  If a tuple, detuple it
    if len(range) == 1 and isinstance(range[0], tuple):
        range = range[0]
        
    # if range[0] is a list or a set, then ignore range[1]
    if isinstance(range[0], list) or isinstance(range[0], set):
        if not value in range[0]:
            iset = range[0]
            pset = format(iset)[0:maxwidth]
            if iset != pset: 
                pset += '...'
            print ('{0:06x}:\tRange Error: {1}={2} ({3})'.format(start, fieldname, value, pset), file=rangeerr)
    
    # If range[0] is int or float then it means min, and range[1] is max, else ignore
    elif isinstance(range[0], int) or isinstance(range[0], float):
        if value < range[0] or value > range[1]:
            print ('{0:06x}:\tRange Error: {1}={2} ({3}..{4})'.format(start, fieldname, value,range[0],range[1]), file=rangeerr)

    else:
        raise invalidRangeCheck('start={0}, fieldname={1}, range={2}'.format(start, fieldname, origrange))

# ###############################################################################
class Struct:
    def __init__(self,
#	deal with fact that version 2.4 doesn't support struct.Struct class
        format):
#			format to pass into struct.xxx methods
# ###############################################################################

        self.format = format

# ###############################################################################
#	expose two methods
    def pack (self, *v):
        return struct.pack(self.format, *v)
    def unpack (self, string):
        return struct.unpack(self.format, string)


tf = '%Y-%m-%d-%H:%M:%S'

class unexpectedEOF(Exception): pass
class parameterError(Exception): pass

# ###############################################################################
class bfile:
# ###############################################################################
    """
    Create an object of this class to work with an associated with a binary file
    
    :param byteorder: indicate byte order character, per struct fmt param '>' is big-endian (default), '<' is little-endian, see http://docs.python.org/release/2.6.6/library/struct.html#byte-order-size-and-alignment
    """
# ###############################################################################
    def __init__(self, byteorder='>'):
# ###############################################################################

        self.__currloc = 0
        
        allowedbyteorder = ['>','<','@','=','!']
        if byteorder not in allowedbyteorder:
            raise parameterError('byteorder must be one of {0}'.format(allowedbyteorder))

        # some handy structures
        self.sint8 = Struct (byteorder+'b')
        self.uint8 = Struct (byteorder+'B')
        self.sint16 = Struct (byteorder+'h')
        self.uint16 = Struct (byteorder+'H')
        self.sint32 = Struct (byteorder+'l')
        self.uint32 = Struct (byteorder+'L')
        self.sint64 = Struct (byteorder+'q')
        self.uint64 = Struct (byteorder+'Q')
        self.float32 = Struct (byteorder+'f')
        self.float64 = Struct (byteorder+'d')

        self.ustruct = {1:self.uint8, 2:self.uint16, 4:self.uint32, 8:self.uint64}
        self.sstruct = {1:self.sint8, 2:self.sint16, 4:self.sint32, 8:self.sint64}
        self.fstruct = {4:self.float32, 8:self.float64}

        
# ###############################################################################
    def open (self, filename, mode):
# ###############################################################################
        """
        open file object

        :param filename: name of file to open
        :param mode: mode to open file - must include 'b' to support binary read / write
        """

        self.file = open(filename,mode)
        self.__currloc = 0

# ###############################################################################
    def close (self): 
# ###############################################################################
        """
        close the file object
        """
        self.file.close()

# ###############################################################################
    def uget (self, numbytes):
# ###############################################################################
        """
        get unsigned bytes

        
        :param numbytes: number of bytes to read from the file
        :rtype: str with bytes from file
        """

        self.__currloc += numbytes
        data = self.file.read(numbytes)
        if len(data) < numbytes: raise unexpectedEOF
        (rtnval,) = self.ustruct[numbytes].unpack(data)
        return rtnval

# ###############################################################################
    def sget (self, numbytes):
# ###############################################################################
        """
        get signed bytes

        
        :param numbytes: number of bytes to read from the file
        :rtype: str with bytes from file
        """

        self.__currloc += numbytes
        data = self.file.read(numbytes)
        if len(data) < numbytes: raise unexpectedEOF
        (rtnval,) = self.sstruct[numbytes].unpack(data)
        return rtnval

# ###############################################################################
    def fget (self, numbytes):
# ###############################################################################
        """
        get floating point bytes

        
        :param numbytes: number of bytes to read from the file
        :rtype: str with bytes from file
        """

        self.__currloc += numbytes
        data = self.file.read(numbytes)
        if len(data) < numbytes: raise unexpectedEOF
        (rtnval,) = self.fstruct[numbytes].unpack(data)
        return rtnval

# ###############################################################################
    def uput (self, numbytes, buffer):
# ###############################################################################
        """
        put unsigned bytes

        :param numbytes: number of bytes to write to the file
        :param buffer: str with bytes to be written to file
        """

        self.__currloc += numbytes
        self.file.write(self.ustruct[numbytes].pack(buffer))

# ###############################################################################
    def sput (self, numbytes, buffer):
# ###############################################################################
        """
        put signed bytes

        :param numbytes: number of bytes to write to the file
        :param buffer: str with bytes to be written to file
        """

        self.__currloc += numbytes
        self.file.write(self.sstruct[numbytes].pack(buffer))

# ###############################################################################
    def fput (self, numbytes, buffer):
# ###############################################################################
        """
        put floating point bytes

        :param numbytes: number of bytes to write to the file
        :param buffer: str with bytes to be written to file
        """

        self.__currloc += numbytes
        self.file.write(self.fstruct[numbytes].pack(buffer))

# ###############################################################################
    def get (self, numbytes):
# ###############################################################################
        """
        get raw bytes

        
        :param numbytes: number of bytes to read from the file
        :rtype: str with bytes from file
        """

        self.__currloc += numbytes
        rtnval = self.file.read(numbytes)
        return rtnval

# ###############################################################################
    def put (self, buffer):
# ###############################################################################
        """
        put raw bytes

        :param buffer: str with bytes to be written to file
        """

        numbytes = len(buffer)
        self.__currloc += numbytes
        self.file.write(buffer)

# ###############################################################################
    def seek (self, loc):
# ###############################################################################
        """
        seek a new location
        
        :param loc: location to put the read or write pointer for the file
        :rtype: int with the previous location
        """
        
        prevloc = self.file.tell()
        self.file.seek(loc)
        self.__currloc = loc     
        return prevloc

# ###############################################################################
    def currloc (self):
# ###############################################################################
        """
        return the current location in the file
        
        :rtype: int with current location
        """

        return self.__currloc

# ###############################################################################
# Keep these constants, exceptions and functions with brecord class
# ###############################################################################
(NDXNDX,LENNDX,SIGNNDX,RNGNDX) = list(range(4))      # LENNDX points at an int
(NDXNDX,CNTFNDX,STRUCNDX,HNDLRNDX) = list(range(4))    # CNTFNDX points at a string
(U,TS,S,F) = list(range(4))   # U = unsigned, TS = timestamp, S = signed, F = float

class invalidSubrecCount(Exception): pass
class invalidStruct(Exception): pass
class invalidInputKeyNotFound(Exception): pass
class invalidCount(Exception): pass

# ###############################################################################
def exprfilter(exprstr, varprefix, varsuffix):
# ###############################################################################
    """
    filter an expression string with certain operators, to an updated expression
    string, that has varprefix before each variables (i.e., any part of the string
    which isn't an operator) and has varsuffix after each variable
    
    valid operators are ( ) * +
    
    :param exprstr: expression string, e.g., '((XXX+YY)*Z)'
    :param varprefix: prefix to put on variables within exprstr, e.g., 'x['
    :param varsuffix: suffix to put on variables within exprstr, e.g., ']'
    :rtype: string with variables updated based on varprefix, varsuffix, e.g., '((x[XXX]+x[YY])*x[Z])'.  This is suitable for call to :func:`eval()`
    """
    
    
    p = re.compile('[()*+]')
    
    # verify the expression has the correct number of parenthesis
    operators = p.findall(exprstr)
    parencheck = 0
    for o in operators:
        if o == '(':
            parencheck += 1
        elif o == ')':
            parencheck -= 1
    if parencheck != 0:
        raise invalidCount('unmatched parenthesis in {0}'.format(exprstr))
    
    # create iterator which demarcates the operators vs. the variables
    # incrementally add operators and updated variables to expression
    iter = p.finditer(exprstr)
    retval = ''
    lastend = 0
    for m in iter:
        if lastend == m.start():
            retval += m.group() # add the operator for this group
        else:
            retval += varprefix + exprstr[lastend:m.start()] + varsuffix
            retval += m.group() # add the operator for this group
        lastend = m.end()
    if lastend != len(exprstr):
        retval += varprefix + exprstr[lastend:len(exprstr)] + varsuffix
    
    return retval

# ###############################################################################
class brecord():
# ###############################################################################
    """
    Generic binary record
    
    :param recstruct: dictionary containing binary record structure
    
        where
            recstruct = 
                {'fieldname':(index,length,sign[,range]), ...}            "scalar field"
                
                or  {'recordname':[index,countfield,structure,None], ...}  "record field" (4th entry of list must be "None")
                
                    index = incremented from 0 for each field -- this needs to be maintained because python 2.6 does not support ordered dictionaries
                    
                    length = length in bytes of field
                    
                    sign = bfile.U - unsigned, bfile.TS - timestamp, bfile.S - signed, bfile.F - float
                    
                    range = optional tuple with two values for (min,max), or list with possible values for [val1,val2, ..., valn]
                    
                    countfield = 'fieldname' of field which has the count for the number of records in this list
                    
                    structure = recstruct pointing to sub-record (recursive)
                    
                    None = must be 'None' (see example) - placeholder for internal use of :class:`bfile.brecord`

    :param dorangecheck: if True, upon :meth:`get`, perform range check with errors to rangeerr, default False
    :param rangeerr: file handle, default sys.stdout
    """
    
    # ###############################################################################
    def __init__(self, recstruct, dorangecheck=False, rangeerr=sys.stdout, rangeignore=[]):
    # ###############################################################################
        self.recstruct = recstruct
        
        fieldlist = []
        r = self.recstruct  # local convenience
        for field in list(r.keys()):
            fieldlist += [(r[field][NDXNDX], (field, r[field]))]  # pull out the field index, to sort
        fieldlist.sort()
        
        self.ordered = [fieldent[1] for fieldent in fieldlist]
        self._dorangecheck = dorangecheck
        self._rangeerr = rangeerr
        self._rangeignore = rangeignore
 
    # ###############################################################################
    def get(self, BF, kwargs):
    # ###############################################################################
        """
        Retrieves a record from a binary file.
        Perform range check on fields if :class:`brecord` instantiated with dorangecheck=True
        
        :param BF: bfile object opened in read mode
        :param kwargs: {'errdisplayfields':list} of fields which should be displayed upon unexpectedEOF exception
        :rtype: dictionary with the data from the record, with appropriate field keys.  Subrecords are ordered lists with these dictionaries.
        """
        
        # dump requested fields if this dictionary has them
        def _dumperrfields(d,fields):
            dumpstr = ''
            for f in list(d.keys()):
                if f in fields:
                    dumpstr += '{0}={1}, '.format(f,d[f])
            if len(dumpstr) > 0:
                print('Unexpected end of file within record having {0}'.format(dumpstr[0:-2]))
            
        retval = {}
        
        # get the error display fields, default []
        errdisplayfields = kwargs.get('errdisplayfields',[])
            
        # Look at each field data in the ordered list.
        for f in self.ordered:
            field = f[0]
            length = f[1][LENNDX]   # take a chance -- it might be the length if a scalar field
            start = BF.currloc()    # for rangecheck, need to know the number of bytes processed in the file
            
            # For scalar fields, just get from the file
            if isinstance(length, int):
                try:
                    if f[1][SIGNNDX] == U or f[1][SIGNNDX] == TS:
                        retval[field] = BF.uget(length)
                    elif f[1][SIGNNDX] == S:
                        retval[field] = BF.sget(length)
                    elif f[1][SIGNNDX] == F:
                        retval[field] = BF.fget(length)
                    else:
                        raise invalidStruct(f)
                    
                    # Perform range check if desired, and if field has a range to be checked, and if we're not supposed to ignore this field
                    dorangecheck = self._dorangecheck and len(f[1]) >= RNGNDX+1 and not field in self._rangeignore
                    if dorangecheck: 
                        rangecheck(start,field,retval[field],f[1][RNGNDX],rangeerr=self._rangeerr)
           
                except unexpectedEOF:
                    print('Unexpected end of file in {0}'.format(field))
                    _dumperrfields (retval, errdisplayfields)
                    raise
                    
            # If a "record field", create a handler if necessary, 
            # then get the appropriate number of those records (note this is recursive)
            else:
                # fill list with entries containing subrecord
                retval[field] = []
                countfield = f[1][CNTFNDX]

                # evaluate the expression indicated by the countfield
                # assumes count field was seen before the records
                count = eval(exprfilter(countfield,'retval["','"]'))

                # create a handler if necessary
                if f[1][HNDLRNDX] == None:
                    recstruct = f[1][STRUCNDX]
                    handler = brecord(recstruct,self._dorangecheck, self._rangeerr, self._rangeignore)
                    f[1][HNDLRNDX] = handler
                handler = f[1][HNDLRNDX]
                
                # get the records from the file.  store in a list
                for subrec in range(count):
                    try:
                        retval[field] += [handler.get(BF,kwargs)]
                    except unexpectedEOF:
                        print('Unexpected end of file in {0}'.format(field))
                        _dumperrfields (retval, errdisplayfields)
                        raise
            
        return retval
        
    # ###############################################################################
    def put(self, BF, record):
    # ###############################################################################
        """
        Inserts a record into a binary file
        
        :param BF: bfile object opened in write mode
        :param record: dictionary with the data from the record, with appropriate field keys.  Subrecords are ordered lists with these dictionaries.
        """
        
        # Look at each field data in the ordered list.
        for f in self.ordered:
            field = f[0]
            
            # unfortunately this doesn't provide a lot of information
            if not field in list(record.keys()):
                raise invalidInputKeyNotFound(field)
                
            length = f[1][LENNDX]  # take a chance -- it might be the length if a scalar field
            # For scalar fields, just put to the file
            if isinstance(length, int):
                if f[1][SIGNNDX] == U or f[1][SIGNNDX] == TS:
                    BF.uput(length, record[field])
                elif f[1][SIGNNDX] == S:
                    BF.sput(length, record[field])
                elif f[1][SIGNNDX] == F:
                    BF.fput(length, record[field])
                else:
                    raise invalidStruct(f)
            
            # If a "record field", create a handler if necessary, 
            # then put the appropriate number of those records (note this is recursive)
            else:
                # fill list with entries containing subrecord
                countfield = f[1][CNTFNDX]

                # evaluate the expression indicated by the countfield
                # assumes count field was seen before the records
                count = eval(exprfilter(countfield,'record["','"]'))

                # create a handler if necessary
                if f[1][HNDLRNDX] == None:
                    recstruct = f[1][STRUCNDX]
                    handler = brecord(recstruct)
                    f[1][HNDLRNDX] = handler
                handler = f[1][HNDLRNDX]
                if len(record[field]) != count:
                    raise invalidSubrecCount('{0}, count={1}, #subrecs={2}'.format(field, count, len(record[field])))
                
                # put the records to the file, from the list
                for subrec in record[field]:
                    handler.put(BF, subrec)

    # ###############################################################################
    def gettsfields(self):
    # ###############################################################################
        """
        Get timestamp fields.  Useful in case timestamps need to be shifted in time.
        
        :rtype: parallel dictionary to recstruct (see :class:`bfile.brecord`), which has True value for fields that are of type TS (timestamp)
        """
    
        retval = {}

        # Look at each field data in the ordered list.
        for f in self.ordered:
            field = f[0]
            
            length = f[1][LENNDX]  # take a chance -- it might be the length if a scalar field
            # For scalar fields, just put to the file
            if isinstance(length, int):
                if f[1][SIGNNDX] == TS:
                    retval[field] = True    # flag timestamp fields
                else:
                    retval[field] = False
            
            # If a "record field", create a handler if necessary, 
            # then put the appropriate number of those records (note this is recursive)
            else:
                # create a handler if necessary
                if f[1][HNDLRNDX] == None:
                    recstruct = f[1][STRUCNDX]
                    handler = brecord(recstruct)
                    f[1][HNDLRNDX] = handler
                handler = f[1][HNDLRNDX]
                
                # get the time stamps for the subrecord
                retval[field] = handler.gettsfields()
        
        return retval
        
    # ###############################################################################
    def getcountfields(self):
    # ###############################################################################
        """
        Get count fields.  Useful to automatically process a brecord file
        
        :rtype: parallel dictionary to recstruct (see :class:`bfile.brecord`), which has False value for scalars and {'count':countfield,'rec':subrecord} dictionary for subrecords
        """
    
        retval = {}

        # Look at each field data in the ordered list.
        for f in self.ordered:
            field = f[0]
            
            length = f[1][LENNDX]  # take a chance -- it might be the length if a scalar field
            # For scalar fields, just put to the file
            if isinstance(length, int):
                retval[field] = False
            
            # If a "record field", create a handler if necessary, 
            # then put the appropriate number of those records (note this is recursive)
            else:
                # create a handler if necessary
                if f[1][HNDLRNDX] == None:
                    recstruct = f[1][STRUCNDX]
                    handler = brecord(recstruct)
                    f[1][HNDLRNDX] = handler
                handler = f[1][HNDLRNDX]
                
                # get this count field and the count fields for the subrecord
                retval[field] = {'count':f[1][CNTFNDX],'rec':handler.getcountfields()}
        
        return retval
        
    # ###############################################################################
    def getfieldorder(self):
    # ###############################################################################
        """
        Get fields in order.  Useful to automatically process a brecord file.
        
        :rtype: list which is parallel to recstruct (see :class:`bfile.brecord`), ordered list of fieldnames, or (fieldname, [subfields...]) recursively
        """
    
        retval = []

        # Look at each field data in the ordered list.
        for f in self.ordered:
            field = f[0]
            
            length = f[1][LENNDX]  # take a chance -- it might be the length if a scalar field
            # For scalar fields, just put to the file
            if isinstance(length, int):
                retval += [field]
            
            # If a "record field", create a handler if necessary, 
            # then put the appropriate number of those records (note this is recursive)
            else:
                # create a handler if necessary
                if f[1][HNDLRNDX] == None:
                    recstruct = f[1][STRUCNDX]
                    handler = brecord(recstruct)
                    f[1][HNDLRNDX] = handler
                handler = f[1][HNDLRNDX]
                
                # get this count field and the count fields for the subrecord
                retval += [(field, handler.getfieldorder())]
        
        return retval
        



###########################################################################################
#
#       Date            Author          Reason
#       ----            ------          ------
#       12/14/17        Lou King        Create
#
#   Copyright 2015 Lou King.  All rights reserved
###########################################################################################

# standard
import xml.etree.ElementTree as ET
import math

###########################################################################################
class LatLng():
###########################################################################################
    #----------------------------------------------------------------------
    def __init__(self, file, filetype, getelev=False):
    #----------------------------------------------------------------------
        if filetype not in ['gpx', 'kml']:
            raise 'invalid type: ' + filetype
        self.filetype = filetype
        self.file = file
        self.getelev = getelev
        self._xml = ET.parse(self.file)

        # set namespace, assume ns is always the same as root
        self.root = self._xml.find('.')
        if self.root.tag[0] == '{':
            self.namespace = '{{{}}}'.format(self.root.tag[1:].split('}')[0])
        else:
            self.namespace = ''

        self._points = self._parse()

    #----------------------------------------------------------------------
    def getpoints(self):
    #----------------------------------------------------------------------
        return self._points  
  
    #----------------------------------------------------------------------
    def _parse(self):
    #----------------------------------------------------------------------
        if (self.filetype == 'kml'):
            return self._parsekml()
        elif (self.filetype == 'gpx'):
            return self._parsegpx()
  
    #----------------------------------------------------------------------
    def _parsekml(self):
    #----------------------------------------------------------------------
        coordstag = './{ns}Document/{ns}Placemark/{ns}LineString/{ns}coordinates'.format(ns=self.namespace)

        coordinates = self.root.find(coordstag)
        coords = coordinates.text.strip().split(' ')
        points = []
        for coord in coords:
            latlng = coord.split(',')
            point = [float(latlng[1]), float(latlng[0])]
            if self.getelev and len(latlng) >= 3: 
                point.append(float(latlng[2]))
            points.append(point)
        return points
  
    #----------------------------------------------------------------------
    def _parsegpx(self):
    #----------------------------------------------------------------------
        trkpttag = './{ns}trk/{ns}trkseg/{ns}trkpt'.format(ns=self.namespace)

        xmlpts = self.root.findall(trkpttag)
        points = []
        for xmlpt in xmlpts:
            point = [
                     float(xmlpt.attrib['lat']),
                     float(xmlpt.attrib['lon'])
                    ]
            if self.getelev:
                ele = xmlpt.find("{ns}ele".format(ns=self.namespace))
                if ele is not None: 
                    point.append(float(ele.text))
            points.append(point)
        return points

###########################################################################################
class GeoDistance():
###########################################################################################
    '''
    calculation functions which depend on Earth radius

    :param R: radius of Earth at nominal desired latitude
    '''

    #----------------------------------------------------------------------
    def __init__(self, R):
    #----------------------------------------------------------------------
        self.R = R

    # compute haversine distance with elevation accounted for
    # https://stackoverflow.com/questions/14560999/using-the-haversine-formula-in-javascript
    # https://math.stackexchange.com/questions/2075092/haversine-formula-that-includes-an-altitude-parameter
    #----------------------------------------------------------------------
    def haversineDistance(self, coords1, coords2, isMiles=True):
    #----------------------------------------------------------------------
        '''
        calculate the distance between two lat,lng,ele coordinates

        Note: the ele item in the coordinate tuple is optional

        :param coords1: [lat, lng, ele] or [lat, lng] lat,lng dec degrees, ele meters
        :param coords2: [lat, lng, ele] or [lat, lng]
        :param isMiles: if True return miles, if False return km
        :rtype: distance between the points
        '''
        lat1 = coords1[0]
        lon1 = coords1[1]
        ele1 = coords1[2] if len(coords1)>=3 else 0.0 # units feet or meters

        lat2 = coords2[0]
        lon2 = coords2[1]
        ele2 = coords2[2] if len(coords2)>=3 else 0.0

        x1 = lat2 - lat1
        dLat = math.radians(x1)
        x2 = lon2 - lon1
        dLon = math.radians(x2)
        a = (math.sin(dLat / 2) * math.sin(dLat / 2) +
             math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) *
             math.sin(dLon / 2) * math.sin(dLon / 2))
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        d = self.R * c

        if isMiles:
            d /= 1.609344

        # account for elevation. ok if difference is negative because squaring
        if isMiles:
            ele1 /= 5280
            ele2 /= 5280
        else:
            ele1 /= 1000
            ele2 /= 1000
        
        de = ele2 - ele1
        d = math.sqrt(d*d + de*de)

        return d

    # from https://gis.stackexchange.com/questions/157693/getting-all-vertex-lat-long-coordinates-every-1-meter-between-two-known-points
    #----------------------------------------------------------------------
    def getDestinationLatLng(self, coord, azimuth, distance):
    #----------------------------------------------------------------------
        '''
        returns the lat and lng of destination point 
        given the start lat, long, azimuth, and distance

        :param coord: [lat,lng]
        :param azimuth: direction in degrees
        :param distance: distance in meters
        :rtype: [lat, lng]
        '''
        brng = math.radians(azimuth) 
        d = distance/1000 
        lat1 = math.radians(coord[0]) 
        lon1 = math.radians(coord[1]) 
        lat2 = math.asin(math.sin(lat1) * math.cos(d/self.R) + math.cos(lat1) * math.sin(d/self.R) * math.cos(brng))
        lon2 = lon1 + math.atan2(math.sin(brng) * math.sin(d/self.R) * math.cos(lat1), math.cos(d/self.R) - math.sin(lat1) * math.sin(lat2))

        lat2 = math.degrees(lat2)
        lon2 = math.degrees(lon2)
        return [lat2, lon2]

# from https://gis.stackexchange.com/questions/157693/getting-all-vertex-lat-long-coordinates-every-1-meter-between-two-known-points
#----------------------------------------------------------------------
def calculateBearing(coord1, coord2):
#----------------------------------------------------------------------
    '''
    calculates the azimuth in degrees from start point to end point

    :param coord1: start point [lat, lng]
    :param coord2: end point [lat, lng]
    :rtype: azimuth in degrees
    '''
    startLat = math.radians(coord1[0])
    startLong = math.radians(coord1[1])
    endLat = math.radians(coord2[0])
    endLong = math.radians(coord2[1])

    dLong = endLong - startLong
    dPhi = math.log(math.tan(endLat/2.0+math.pi/4.0)/math.tan(startLat/2.0+math.pi/4.0))
    if abs(dLong) > math.pi:
        if dLong > 0.0:
            dLong = -(2.0 * math.pi - dLong)
        else:
            dLong = (2.0 * math.pi + dLong)

    bearing = (math.degrees(math.atan2(dLong, dPhi)) + 360.0) % 360.0;
    return bearing


#----------------------------------------------------------------------
def elevation_gain(elevations, isMiles=True, upthreshold=8, downthreshold=8, debug=False):
#----------------------------------------------------------------------
    '''
    calculate elevation gain over a series of elevation points

    NOTE: thresholds of 8 meters approximately matches strava

    :param elevations: list of elevation points in meters
    :param isMiles:  if True return feet, if False return m
    :param upthreshold: threshold of increase when to decide climbing, meters
    :param downthreshold: threshold of decrease when to decide descending, meters
    :param debug: if True, return tuple with gain, debuginfo
    :rtype: gain[, debuginfo] - gain in meters or feet depending on isMiles
    '''
    # highup keeps track of the highest elevation reached while climbing
    # lowdown keeps track of the lowest elevation reached while descending
    # decision is made to switch from climbing to descending and visa versa
    # when threshold is passed in appropriate direction
    if debug: debuginfo = ['ele,state,highup,lowdown,totclimb\n']
    state = 'unknown'
    thisel = elevations[0]
    highup   = thisel
    lowdown  = thisel
    totclimb = 0.0
    if debug: debuginfo.append('{},{},{},{},{}\n'.format(thisel,state,highup,lowdown,totclimb))
    for thisel in elevations[1:]:
        if state == 'unknown':
            if thisel >= highup + upthreshold:
                state = 'climbing'
                highup = thisel
            elif thisel <= lowdown - downthreshold:
                state = 'descending'
                lowdown = thisel

        elif state == 'climbing':
            if thisel > highup:
                highup = thisel

            elif thisel <= highup - downthreshold:
                state = 'descending'
                totclimb += highup - lowdown
                lowdown = thisel

        elif state == 'descending':
            if thisel < lowdown:
                lowdown = thisel

            elif thisel >= lowdown + upthreshold:
                state = 'climbing'
                highup = thisel

        if debug: debuginfo.append('{},{},{},{},{}\n'.format(thisel,state,highup,lowdown,totclimb))

    # may need to add last climb in
    if state == 'climbing':
        totclimb += highup - lowdown
        if debug:
            debuginfo.pop()
            debuginfo.append('{},{},{},{},{}\n'.format(thisel,state,highup,lowdown,totclimb))

    # convert to feet
    if isMiles:
        totclimb = (totclimb/1609.344) * 5280
    
    if debug: return totclimb, debuginfo

    return totclimb





# adapted from https://gist.github.com/mattupstate/9257466

from flask import Flask, Blueprint
from flask.views import MethodView


class ApiResource(MethodView):
    endpoint = None
    url_prefix = None
    url_rules = {}

    @classmethod
    def as_blueprint(cls, name=None):
        name = name or cls.endpoint
        bp = Blueprint(name, cls.__module__, url_prefix=cls.url_prefix)
        for endpoint, options in list(cls.url_rules.items()):
            url_rule = options
            methods = ('GET',)
            defaults = {}
            if len(options) == 2:
                url_rule, methods = options
            elif len(options) == 3:
                url_rule, methods, defaults = options
            bp.add_url_rule(url_rule, endpoint=endpoint, methods=methods,
                            defaults=defaults, view_func=cls.as_view(endpoint))
        return bp


class MyResource(ApiResource):
    endpoint = 'my'
    url_prefix = '/hello'
    url_rules = {
        'index': ['', ('GET',), {'id': None}],
        'select': ['/<id>', ('GET',)]
    }

    def get(self, id):
        if id is None:
            return 'hello'
        return id


if __name__ == '__main__':
    app = Flask(__name__)
    app.register_blueprint(MyResource.as_blueprint())

    assert app.url_map.is_endpoint_expecting('my.index')
    assert app.url_map.is_endpoint_expecting('my.select', 'id')

    with app.test_client() as client:
        assert client.get('/hello').data == 'hello'
        assert client.get('/hello/stuff').data == 'stuff'

    print('Tests passed successfully!')


'''
mailer - send email
================================================
'''
# standard

# pypi
from flask import current_app
from flask_mail import Message

debug = False

#----------------------------------------------------------------------
def sendmail(subject, fromaddr, toaddr, html, text='', ccaddr=None, replytoaddr=None):
#----------------------------------------------------------------------
    '''
    send mail

    :param subject: subject of email
    :param fromaddr: from address to use
    :param toaddr: to address to use, may be list of addresses or comma separated
    :param html: html to send
    :param text: optional text alternative to send
    :param ccaddr: optional cc address to use, may be list of addresses or comma separated
    :param replytoaddr: optional reply_to address to use, may be list of addresses or comma separated
    :returns: response from flask_mail.send
    '''
    current_app.logger.info('sendmail(): from={}, to={}, cc={}, subject="{}"'.format(fromaddr, toaddr, ccaddr, subject))

    # get current app's mailer
    mail = current_app.extensions.get('mail')

    message = Message(
        sender=fromaddr,
        recipients=toaddr if isinstance(toaddr, list) else [a.strip() for a in toaddr.split(',')],
        cc=ccaddr if not ccaddr or isinstance(ccaddr, list) else [a.strip() for a in ccaddr.split(',')],
        reply_to=replytoaddr if not replytoaddr or isinstance(replytoaddr, list) else [a.strip() for a in replytoaddr.split(',')],
        subject=subject,
        html=html,
        body=text,
    )

    mail.send(message)

    if debug: current_app.logger.debug('sendmail(): message.html={}'.format(message.html))
